from functools import partial
from typing import Optional

from django.http import HttpRequest

from .settings import django_settings, AnyCaseSettings
from .utils import json_loads, is_json_possible
from ..parser import CaseFormatParser


def parse_header(key: str, request: HttpRequest) -> Optional[str]:
    return request.headers.get(key)


def parse_query(key: str, request: HttpRequest) -> Optional[str]:
    return request.GET.get(key) or request.POST.get(key)


def parse_body(key: str, request: HttpRequest) -> Optional[str]:
    if is_json_possible(request):
        try:
            return json_loads(request).get(key)
        except ValueError:
            pass


def django_case_format_parser(settings: AnyCaseSettings) -> CaseFormatParser:
    parsers = []

    if settings.HEADER_KEY is not None:
        parsers.append(partial(parse_header, settings.HEADER_KEY))
    if settings.QUERY_KEY is not None:
        parsers.append(partial(parse_query, settings.QUERY_KEY))
    if settings.BODY_KEY is not None:
        parsers.append(partial(parse_body, settings.BODY_KEY))

    return CaseFormatParser(parsers)


case_format_parser = django_case_format_parser(django_settings)



import importlib
from typing import Any

import django.conf
from django.utils.functional import SimpleLazyObject

DEFAULTS = {
    'HEADER_KEY': 'Accept-Json-Case',
    'QUERY_KEY': None,
    'BODY_KEY': None,
    'JSON_MODULE': 'json',
    'CONVERT_INPUT_JSON': True
}

IMPORT_MODULES = ['JSON_MODULE']

__all__ = ['AnyCaseSettings', 'any_case_settings', 'django_settings']


class AnyCaseSettings:
    def __init__(self, settings: dict = None, defaults: dict = None) -> None:
        self._defaults = (defaults or DEFAULTS).copy()
        self._settings = {**self._defaults, **settings.copy()}

    def __getattr__(self, attr: str) -> Any:
        if attr not in self._defaults:
            raise AttributeError(f'Unknown setting: {attr!r}')

        if attr in IMPORT_MODULES:
            value = SimpleLazyObject(
                lambda: importlib.import_module(self._settings[attr])
            )
        else:
            value = self._settings[attr]

        setattr(self, attr, value)
        return value

    @property
    def has_convert_key(self) -> bool:
        return any([self.HEADER_KEY, self.QUERY_KEY, self.BODY_KEY])


def any_case_settings(settings: Any) -> AnyCaseSettings:
    return AnyCaseSettings(
        settings=getattr(settings, 'ANY_CASE', {}),
        defaults=DEFAULTS,
    )


django_settings = SimpleLazyObject(
    lambda: any_case_settings(django.conf.settings)
)



from typing import Union

from django.http import HttpRequest, HttpResponse
from django.utils.functional import SimpleLazyObject

from .settings import django_settings

json = django_settings.JSON_MODULE


def json_loads(request: HttpRequest) -> Union[dict, list]:
    if hasattr(request, 'json'):
        return request.json

    return json.loads(request.body)


def is_json_possible(request: HttpRequest) -> bool:
    return bool(
            request.method not in {'GET', 'HEAD'} and
            request.content_type == 'application/json' and
            request.body
    )


def is_json_content(response: HttpResponse) -> bool:
    return bool(
            response.get('Content-Type') == 'application/json' and
            response.content
    )



import re
from copy import deepcopy

from typing import TypeVar

__snake_first = re.compile(r'([^_])([A-Z][a-z]+)')
__snake_second = re.compile(r'([a-z])([A-Z])')
__snake_sub = r'\1_\2'


def to_snake_case(word: str) -> str:
    word = __snake_first.sub(__snake_sub, word)
    word = __snake_second.sub(__snake_sub, word)
    return word.lower()


__camel_first = re.compile(r'_([\w])([^_]+)')


def __camel_first_sub(match):
    return match.group(1).upper() + match.group(2).lower()


__camel_second = re.compile(r'([A-Z]+)([A-Z][a-z])')


def __camel_second_sub(match):
    return match.group(1).lower() + match.group(2)


def to_camel_case(word: str) -> str:
    word = __camel_first.sub(__camel_first_sub, word)
    word = __camel_second.sub(__camel_second_sub, word)
    return word[0].lower() + word[1:]


T = TypeVar('T')


def converts_keys(data: T, *, case='snake', inplace=False) -> T:
    if case == 'camel':
        formatter = to_camel_case
    elif case == 'snake':
        formatter = to_snake_case
    else:
        raise ValueError('Invalid case type, use `snake` or `camel`')

    if not inplace:
        data = deepcopy(data)

    queue = [data]

    while queue:
        element = queue.pop()

        if isinstance(element, list):
            queue.extend(element)

        elif isinstance(element, dict):
            for key in list(element.keys())[:]:
                value = element[formatter(key)] = element.pop(key)

                if isinstance(value, list):
                    queue.extend(value)
                elif isinstance(value, dict):
                    queue.append(value)

    return data



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['any_case', 'any_case.contrib', 'any_case.contrib.django']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'any-case',
    'version': '0.1.1',
    'description': 'Snake/Camle case converter with django and rest_framework integration',
    'long_description': ".. highlight:: python\n\nany_case\n========\n\nWhen developing a web application, you often have to choose the case format (snake_case/camelCase)\nwill be used for the input and the output json data. If the backend is written in Python,\nthe easiest option is to choose snake_case. On the one hand, this is good, because we have\nan obvious consistency, but on the other hand, the main consumers of the API (mobile and browser)\nprefer to use camelCase.\n\nIf we care about our consumers, we will change the case format to camelCase. And it's good if we think\nabout it in advance, before publishing API. But if this happens in an existing application,\nit becomes much more difficult to do so because there are consumers using the existing API.\nWe have two options here:\n\n- introduce a new version of the api\n- send data in two cases at once\n\nThe second option increases the size of the data, and the first option forces us to support two\nversions of the API without serious need. These options complicate the support and development of API.\n\nThings get a little more complicated when we have consumers who natively use snake_case,\nand using camelCase for them is not an option at all.\n\nAs you can see, the use of one notation does not improve, but may worsen the situation.\n\nBut why do we have to choose for the customer which case format they use?\nWhy customers do not send and not receive data in the format, which would be more convenient to them?\n\n\nThat's what this library is for. Consumers choose in what case they expect the data.\nIf they specified an incorrect case or made a request without specifying a case,\nthe data will be given as is.\n\nThis approach allows existing consumers to work without changes with existing API, and for those\nconsumers who want to use a single format, allows granular rewriting of the application.\n\nInstallation\n============\n::\n\n    pip install any_case\n\nUsage\n=====\nFor converting dict or list use ``converts_keys`` function::\n\n    >>> from any_case import converts_keys\n    >>> data = {'camelCaseKey': 'value'}\n    >>> converts_keys(data, case='snake')\n    {'camel_case_key': 'value'}\n    >>> data = {'snake_case': 'camelCase'}\n    >>> converts_keys(data, case='camel')\n    {'snakeCase': 'camelCase'}\n    >>>\n\nFor converting ``any_case`` uses compiled regex and stack for objects traversal.\n\nTo convert text, use ``to_snake_case`` or ``to_camel_case``::\n\n    >>> from any_case import to_snake_case, to_camel_case\n    >>> to_snake_case('snakeCase')\n    'snake_case'\n    >>> to_camel_case('snake_case')\n    'snakeCase'\n\nMore examples see at tests folder.\n\nIntegrations\n============\n\nDjango\n------\n\nFor integration with the Django framework, you need to add ``any_case`` middleware::\n\n    MIDDLEWARE = [\n        'django.contrib.sessions.middleware.SessionMiddleware',\n        ...\n        'any_case.contrib.django.middleware.KeysConverterMiddleware'\n    ]\n\n\nWhen the input data is converted, it is saved in the ``json`` field of request object.\nThat is, you can access to the converted data in the view as follows::\n\n    def view(request):\n        data = request.json\n        ...\n\nrest_framework\n--------------\n\nFor integration with rest_framework, replace default json parser and renderer::\n\n    REST_FRAMEWORK = {\n        'DEFAULT_RENDERER_CLASSES': (\n            'any_case.contrib.rest_framework.AnyCaseJSONParser',\n            ...\n        ),\n        'DEFAULT_PARSER_CLASSES': (\n            'any_case.contrib.rest_framework.AnyCaseJSONRenderer',\n            ...\n        )\n    }\n\n\nSettings\n--------\n``any_case`` has the next default settings::\n\n    ANY_CASE = {\n        'HEADER_KEY': 'Accept-Json-Case',\n        'QUERY_KEY': None,\n        'BODY_KEY': None,\n        'CONVERT_INPUT_JSON': True\n    }\n\nSettings are specified in django ``settings.py``.\n\n``any_case`` can be used for converting input json data to snake_case and for converting\noutput json to snake_case or camelCase. Or only one of the above independently.\n\nYou can specify the format in the header, in the query parameters, or in the json body.\nThe preferred way is the header, because specifying in the query or in the body\nis not always possible. Specifying case format in the body also forces to parse json data that\nmay not be needed at all.\n\n\nDisable converting output data\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n::\n\n    ANY_CASE = {\n        'HEADER_KEY': None,\n        'QUERY_KEY': None,\n        'BODY_KEY': None,\n    }\n\nDisable converting input data\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n::\n\n    ANY_CASE = {\n        'CONVERT_INPUT_JSON': False\n    }\n",
    'author': 'asduj',
    'author_email': 'asduj@ya.ru',
    'url': 'https://github.com/asduj/any_case',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import List, Callable, Any, Optional

Parser = Callable[[Any], Optional[str]]


class CaseFormatParser:
    def __init__(self, parsers: List[Parser]) -> None:
        self.parsers = parsers

    def parse(self, request: Any, raise_exc: bool = False) -> Optional[str]:
        case = None

        for parser in self.parsers:
            case = parser(request)
            if case:
                case = case.lower()
                break

        # TODO Replace set to Enum if the number of cases will be more than two
        if case is None or case not in {'snake', 'camel'}:
            if raise_exc:
                raise ValueError('Did not found known case format')
            else:
                case = None

        return case



from django.http import HttpRequest, HttpResponse
from django.utils.deprecation import MiddlewareMixin

from any_case import converts_keys
from .parser import case_format_parser
from .settings import django_settings, AnyCaseSettings
from .utils import json, json_loads, is_json_possible, is_json_content

__all__ = [
    'InputKeysConverterMixin',
    'KeysConverterMiddleware',
    'OutputKeysConverterMixin',
    'any_case_middleware_class',
]


class InputKeysConverterMixin:
    @staticmethod
    def process_request(request: HttpRequest) -> HttpRequest:
        if is_json_possible(request):
            try:
                json_body = json_loads(request)
                converted = converts_keys(
                    json_body, case='snake', inplace=True,
                )
                request.json = converted
            except ValueError:
                pass

        return request


class OutputKeysConverterMixin:
    @staticmethod
    def process_response(
            request: HttpRequest,
            response: HttpResponse,
    ) -> HttpResponse:
        if is_json_content(response):
            try:
                case = case_format_parser.parse(request, raise_exc=True)
                json_content = json.loads(response.content)
                converted = converts_keys(json_content, case=case, inplace=True)
                response.content = json.dumps(converted)
            except ValueError:
                pass

        return response


def any_case_middleware_class(settings: AnyCaseSettings) -> MiddlewareMixin:
    bases = [MiddlewareMixin]

    if settings.CONVERT_INPUT_JSON:
        bases.append(InputKeysConverterMixin)
    if settings.has_convert_key:
        bases.append(OutputKeysConverterMixin)

    return type('KeysConverterMiddleware', tuple(bases), {})


KeysConverterMiddleware = any_case_middleware_class(django_settings)



from typing import Any, Union
from io import BytesIO

from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer

from .django.parser import case_format_parser
from .django.settings import django_settings
from ..converter import converts_keys


class AnyCaseJSONParser(JSONParser):
    if django_settings.CONVERT_INPUT_JSON:
        def parse(
                self,
                stream: BytesIO,
                media_type: str = None,
                parser_context: dict = None,
        ) -> Any:
            data = super().parse(stream, media_type, parser_context)
            data = converts_keys(data, case='snake', inplace=True)
            return data


class AnyCaseJSONRenderer(JSONRenderer):
    if django_settings.has_convert_key:
        def render(
                self,
                data: Union[dict, list],
                accepted_media_type: str = None,
                renderer_context: dict = None
        ) -> bytes:
            renderer_context = renderer_context or {}
            request = renderer_context['request']

            case = case_format_parser.parse(request)
            if case:
                data = converts_keys(data, case=case, inplace=True)

            return super().render(data, accepted_media_type, renderer_context)
